/* $Id: mga2x64.c,v 1.9 1999/01/29 13:19:19 marcus Exp $
******************************************************************************

   KGI - Matrox Millennium I/II acceleration driver

   Copyright (C) 1997-1998	Jonas Borgstrm	[jonas_b@bitsmart.com]
   Copyright (C) 1998		Marcus Sundberg	[mackan@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Jonas_Borgstrom
#define	DRIVER_NAME	"Matrox Millennium (I/II) accelerator driver"
#define	DRIVER_REV	"$Revision: 1.9 $"

/*
**	KGI includes
*/

#include <kgi/module.h>
#include <kgi/types.h>
#include <kgi/kgi_commands.h>

#include <Matrox/mga_common.h>
#include <Matrox/mga2x64_main.h>
#include <Matrox/mga_common_regs.h>

/*#define NUMPINGPONGPAGES 2*/
#define MGA2X64WWAITFREE() while(mem_in32(STATUS + 2) & 0x01);

/* We don't need the suggest stuff for kgicon */
#undef NEED_SUGGEST_FEAT

#ifdef NEED_SUGGEST_FEAT
const char *sugg_default[KGIGT_LAST][5] = {
	{ "generic-text-16","generic-ramdac",NULL,NULL,NULL },
	{ "generic-text-32","generic-ramdac",NULL,NULL,NULL },
	{ "generic-linear-1","generic-ramdac",NULL,NULL,NULL },
	{ "vendor-IBM-vga-4bpp",NULL,NULL,NULL,NULL},
	{ "generic-linear-8","generic-ioctl","generic-ramdac",NULL,NULL},
	{ "generic-linear-15",NULL,NULL,NULL,NULL},
	{ "generic-linear-16","generic-ioctl",NULL,NULL},
	{ "generic-linear-24","generic-ioctl","generic-gamma",NULL,NULL},
	{ "generic-linear-32","generic-ioctl","generic-gamma",NULL,NULL}
};
#endif /* NEED_SUGGEST_FEAT */

/*
**	KGI module interface
*/
int kgim_accel_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev " DRIVER_REV " loaded");
	/* !!! Check for Chipset and Version here !!! */

	return EOK;
}

void kgim_accel_done(struct kgi_display *dpy)
{
	NOTICE1(DRIVER_NAME " removed");
}
/*
 * Check for a valid graphic mode.
 */
int kgim_accel_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
	enum kgi_tm_cmd *cmd)
{
	kgi_uint chiprounding = 1, tmp = 0, i;
	kgi_uint interleave = (dpy->card->chipset->ramsize > (2 MB));

	/* 
	 * Valid virtual X-resolutions
	 */
	int width[] = { 640, 768, 800, 960, 1024, 1152, 1280,
			1600, 1920, 2048, 0 };

	switch (*cmd) {

	case CMD_CHECK:
		/* 
		 * Check for a valid mode
		*/
		if (mode->request.graphtype == KGIGT_TEXT16 ||
			mode->request.graphtype == KGIGT_4BIT ||
			mode->request.graphtype == KGIGT_15BIT) {

			return EOK;
		}

/*		if (mode->request.graphtype == KGIGT_8BIT &&
			mode->request.virt.x == 320 &&
			mode->request.virt.y == 200) {

			return EOK;
		}*/

		switch (mode->request.graphtype) {

		case KGIGT_8BIT:
			chiprounding = interleave ? 128 : 64;
			break;

		case KGIGT_16BIT:
			chiprounding = interleave ? 64 : 32;
			break;
#if defined(__LITTLE_ENDIAN)
		case KGIGT_24BIT:
			chiprounding = interleave ? 128 : 64;
			break;
#endif
		case KGIGT_32BIT:
			chiprounding = interleave ? 32 : 16;
			break;

		default:
			INTERNAL_ERROR;
		}

		/*
		 * Search for a valid virtual x resolution
		 */
		i = 0;
		do
	  	{
			tmp = width[i];
			if(width[i] == 0) return -E(ACCEL,NOSUP_ALWAYS_CANT);
			i++;
	     	} while((tmp < mode->request.virt.x) || (tmp % chiprounding));
	   	mode->request.virt.x = tmp;

		return EOK;

	default:
		return EOK;
	}
}

/*
 * Setup the accelerator engine.
 */
void kgim_accel_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
	kgiu32 shift = 0;
   
	switch (mode->request.graphtype) {

	case KGIGT_8BIT:
		shift = 0;
#if defined(__BIG_ENDIAN)
		mem_out32(0x01000000, OPMODE);
#endif
		break;

	case KGIGT_16BIT:
		shift = 1;
#if defined(__BIG_ENDIAN)
		mem_out32(0x01010000, OPMODE);
#endif
		break;
#if defined(__LITTLE_ENDIAN)
	case KGIGT_24BIT:
		shift = 3;
		break;
#endif
	case KGIGT_32BIT:
		shift = 2;
#if defined(__BIG_ENDIAN)
		mem_out32(0x01020000, OPMODE);
#endif
		break;

	default:
		return;
	}

	mem_out32(0x7FFF0000, CXBNDRY);
	mem_out32(0x00000000, YTOP);
	mem_out32(mode->request.virt.x*mode->request.virt.y, YBOT);
	mem_out32(0x00000000, FCOL);
	mem_out32(0x00000000, SHIFT);
#if defined(__LITTLE_ENDIAN)
	mem_out32(0x01000000, OPMODE);
#endif
	mem_out32(shift     , MACCESS);
	mem_out32(0x00000000, YDSTORG);
	mem_out32(0xFFFFFFFF, PLNWT);
	mem_out32(mode->request.virt.x, PITCH);
}

/*
***************************************************************************
 Actual acceleration functions

  Most of them are inlined to gain speed.
  As they are only called from one place this doesn't waste any space.
***************************************************************************
*/

/* Set the accelerator's clipping rectangle
 */
static inline void mga2x64_setclip(struct kgi_display *dpy,
				   struct kgi_gc *gc)
{
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);

	mem_out32((gc->cliptl.x & 0x07FF) | (((gc->clipbr.x-1) & 0x07FF) << 16),
		  CXBNDRY);
	mem_out32((dpy->mode.request.virt.x*gc->cliptl.y) & 0x00FFFFFF,
		  YTOP);
	mem_out32((dpy->mode.request.virt.x*(gc->clipbr.y-1)) & 0x00FFFFFF,
		  YBOT);
}

/* Set the accelerator's foreground color
 */
static inline void mga2x64_setfgcolor(struct kgi_display *dpy,
				      unsigned long pixelfg)
{
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);

	switch(dpy->mode.request.graphtype) {
	case KGIGT_8BIT:
		pixelfg &=0xff;   
		pixelfg |= (pixelfg << 8) |
			(pixelfg << 16) | (pixelfg << 24);
		break;
	case KGIGT_16BIT:
		pixelfg &=0xffff;   
		pixelfg |= pixelfg << 16;
		break;
#if defined(__LITTLE_ENDIAN)		
	case KGIGT_24BIT:
		pixelfg &= 0xffffff;
		break;
#endif
	case KGIGT_32BIT:
		pixelfg |= 0xff000000;
		break;
	default:
		return;
	}
	mem_out32(pixelfg, FCOL);
}

/* Update GC if neccesary
 */
static inline void mga2x64_updategc(struct kgi_graphic *gr)
{
	kgi_gc *gc = gr->gc;

	if (gc->version != gc->oldversion) {
		mga2x64_setclip(gr->dev.dpy, gc);
		mga2x64_setfgcolor(gr->dev.dpy, gc->fg_color);
		gc->oldversion = gc->version;
	}
}

/* CopyBox using the accelerator
 */
static inline int mga2x64_copybox(struct kgi_graphic *gr, 
				  struct kgi_cbox *boxp)
{	
#define COPY_LEFT	1
#define COPY_UP		4
	struct kgi_display *dpy = gr->dev.dpy;
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
	kgis32 ar5 = dpy->mode.request.virt.x;
	kgiu32 sgn = 0;
	kgiu32 begin, end;
	kgiu32 x = boxp->x, y = boxp->y, w = boxp->w, h = boxp->h,
		dstx = boxp->x2, dsty = boxp->y2;

	if (dsty > y) {
		sgn |= COPY_UP;
		y += h - 1;
		dsty += h - 1;
		ar5 = - ar5;
	}
	
	begin = end = y * dpy->mode.request.virt.x + x;

	w--;
	if (dstx > x) {
		sgn |= COPY_LEFT;
		begin += w;
	} else {
		end += w;
	}

	mem_out32(BLTMOD_BFCOL | BOP_COPY | SHFTZERO | BITBLT
		  | (sgn ? 0 : SGNZERO), DWGCTL);
	if (sgn) {
		mem_out32(sgn, SGN);
	}
	mem_out32(end, AR0);
	mem_out32(begin, AR3);
	mem_out32(ar5, AR5);
	mem_out32(((dstx + w) << 16) | dstx, FXBNDRY);
	mem_out32((dsty << 16) | h, (kgiu8*)YDSTLEN + ACCEL_GO);
   
	MGA2X64WWAITFREE();
	return EOK;
}

/* Draw a box with the accelerator
 */
static inline int mga2x64_drawbox(struct kgi_graphic *gr, void *iobuf)
{
	struct kgi_display *dpy = gr->dev.dpy;
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
	struct kgi_box *boxp = (struct kgi_box *)iobuf;

	if (boxp->w && boxp->h) {	/* 0 width is not OK! */
		mem_out32(((boxp->x + boxp->w) << 16) | boxp->x, FXBNDRY);
		mem_out32((boxp->y << 16) | boxp->h, YDSTLEN);
		mem_out32(0x000C7804, DWGCTL + ACCEL_GO);
   
		MGA2X64WWAITFREE();
	}
	return EOK;
}

/* Draw a line using the accelerator
 */
static inline int mga2x64_drawline(struct kgi_graphic *gr, void *iobuf)
{
	struct kgi_display *dpy = gr->dev.dpy;
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
	struct kgi_line *linp = (struct kgi_line *)iobuf;
   
	mem_out32(linp->x | (linp->y << 16), XYSTRT);
	mem_out32(linp->x2 | (linp->y2 << 16), XYEND);
	mem_out32(0x040C4803, DWGCTL + ACCEL_GO);
  
	MGA2X64WWAITFREE();
	return EOK;
}


int kgim_accel_command (struct kgi_graphic *gr, kgi_uint cmd, kgi_uint arg)
{
	if (!(gr->dev.state & KGI_MAPPED_ON_DISPLAY) ) {
		/* Don't accel inactive screen! */
	        return -E(ACCEL, NOSUP_MODE_MMAP);
	}

	mga2x64_updategc(gr);

	switch (cmd) { 
	case ACCEL_COPYBOX:
		return mga2x64_copybox(gr, (struct kgi_cbox*) gr->io_buf);
	   
	case ACCEL_DRAWLINE:
		return mga2x64_drawline(gr, gr->io_buf);
	   
	case ACCEL_DRAWBOX:
		return mga2x64_drawbox(gr, gr->io_buf);

#ifdef NEED_SUGGEST_FEAT
	case ACCEL_GETSUGGEST:
	{
		kgi_suggest *sugg;
		const char *name;

		sugg = (kgi_suggest *)(gr->io_buf);
		gr->io_res = gr->io_buf;
      
		if (sugg->id < 0 || sugg->id > 4) {

			return -E(ACCEL, BOGUS);
		}
      
		name = sugg_default[gr->dev.mode.graphtype][sugg->id];
		if (!name) {

			return -E(ACCEL, BOGUS);
		}
/*		if (gr->dev.mode.graphtype == KGIGT_8BIT &&
			gr->dev.mode.virt.x == 320 &&
			gr->dev.mode.virt.y == 200 &&
			strcmp("generic-ioctl", name) == 0) {
		   
			return -E(ACCEL, BOGUS);
		}*/
      
		strcpy(sugg->name, name);
		sugg->args[0] = 0;
		sugg->id++;	/* Next is RAMDAC */
		if (sugg->id >= 4 ||
		    sugg_default[gr->dev.mode.graphtype][sugg->id] == NULL) {

			sugg->id = 0;
		}
		return 0;
	}
#endif /* NEED_SUGGEST_FEAT */

	default: 
	  
	        return -E(ACCEL,NOSUP_ALWAYS_MMAP); 
	}
}
